#include <stdio.h>
#include <syslog.h>
#include <stdlib.h>
#include <signal.h>

#include "pam_smb_queue.h"
#include "cache.h"

char primary[60];
char backup[60];
char domain[60];

void *sighup_handler(int signum);
void *sigusr2_handler(int signum);
void *sigterm_handler(int signum);

int mid;
int debug=1;

/* domain_list *list_of_domains; */
user_list *user_cache;

int main(void)
{
  int key;
  struct sigaction sighup_act, sigusr2_act, sigterm_act;

  openlog("pamsmbd", LOG_PID, LOG_AUTHPRIV);

  /* Allocate memory for cache */
  user_cache = (user_list *) malloc(sizeof(user_list));
  if (user_cache==NULL)
    {
      syslog(LOG_AUTHPRIV | LOG_ERR, "Out of memory ...");
      exit(-1);
    }

  user_cache->head=NULL;
  user_cache->last=NULL;
  
  /*
  list_of_domains = (domain_list *) malloc(sizeof(user_list));
  if (user_cache==NULL)
    {
      syslog(LOG_AUTHPRIV | LOG_ERR, "Out of memory... ");
      exit(-1);
    }
    */
  /* set up signal handlers */
  /* SIGHUP re-read configuration file */
  /* SIGUSR2 dumps cache to /tmp */ 
  /* SIGTERM close msgqueue and ends program */

  sighup_act.sa_handler=sighup_handler;
  sigaction(SIGHUP, &sighup_act, NULL);
  sigusr2_act.sa_handler=sigusr2_handler;
  sigaction(SIGUSR2, &sigusr2_act, NULL);
  sigterm_act.sa_handler=sigterm_handler;
  sigaction(SIGTERM, &sigterm_act, NULL);
  sigaction(SIGINT, &sigterm_act, NULL);
  sigaction(SIGQUIT, &sigterm_act, NULL);

  switch(fork())
	 {
	case 0:
		break;
	case -1:
		printf("error in fork\n");
		exit(-1);
	default:
		exit(0);
	}

  setsid();

  /* Read configuration */
  smb_readpamconf(primary, backup, domain);

  /* Setup the SYSV IPC msg queue */
  mid=setup_queue();

  /* Start reading the queue */
  do_queue(mid);

  closelog();
}

int setup_queue(void)
{
  int mid;
  int key=PAMSMBKEY;
  
  /* if queue exists remove it and re-create it */
  if((mid=msgget(key, 0)) != -1)
    {
      msgctl(mid, IPC_RMID, NULL);
    }
  
  mid=msgget(key, IPC_CREAT | 0600);
  if (mid==-1)
    {
      perror("msgget: ");
      exit(-1);
    }

  return mid;
}

/* function to get queue entries and validate them */  
int do_queue(int mid)
{
  int n;
  int ok=0;
  MESSAGE msg;
  int user_num;
  user_cache_list *loop, *found;
  int cache_status;
  int valid=0;
  int position;
  user_cache_list *newentry;

  while(1)
    {
      ok=0;
      do {
	n=msgrcv(mid, &msg, sizeof(msg), SERVER, 0);
	if (n==-1)
	  {
	    if (errno!=EINTR)
	      {		
		perror("Server: msgrcv");
		msgctl(mid ,IPC_RMID, NULL);
		exit(-1);
	      }
	  }
	else
		ok=1;
      } while (ok==0);

      newentry = (user_cache_list *) malloc(sizeof (user_cache_list));
      if (newentry==NULL)
	{
	  syslog(LOG_AUTHPRIV | LOG_ERR, "pamsmbd: Out of Memory");
	  exit(-1);
	}
      memset(newentry,0, sizeof(user_cache_list));

      strncpy(newentry->nt_user, msg.username, 21);
      strncpy(newentry->ux_user, msg.username, 21);
      strncpy(newentry->password, msg.password, 39);
      newentry->next = NULL;

      pam_map_user(newentry);

      /* Check cache for user code */
      /* Cache will *always* be checked on UNIX username */

      user_num = user_cache->number_of_cached_users;
      position=0;
      ok=0;
      found=NULL;
      if (user_cache->head==NULL)
	{
	  if(debug) syslog(LOG_AUTHPRIV | LOG_DEBUG,"Cache is empty\n");
	  cache_status=CACHE_ENOENT;
	}
      else
	for (loop=user_cache->head; loop!=NULL; loop=loop->next)
	  {
	    if ((strcmp(loop->ux_user, newentry->ux_user)==0) &&
		(strcmp(loop->password, newentry->password)==0))
	      {
	      position++;
	      if ((time(NULL) - loop->usertimeout)<CACHE_TIMEOUT)
		{
		  ok=1;
		  found=loop;
		  cache_status=CACHE_OK;
		  if (debug) syslog(LOG_AUTHPRIV | LOG_DEBUG, "Cached User Ok");
		}
	      else
		{
		  ok=1;
		  found=loop;
		  cache_status=CACHE_ETIME;
		}
	      }
	  else
	    if(ok==0) 
	      cache_status=CACHE_ENOENT;
	  }
      
      if (cache_status!=CACHE_OK)
	{


	  /* if nothing in the map for primary, backup and domain
	     user defaults from conf file */

	  if ((strlen(newentry->pdc_nbname)==0) || (strlen(newentry->nt_domain)==0))
	    {
	      strcpy(newentry->pdc_nbname, primary);
	      strcpy(newentry->bdc_nbname, backup);
	      strcpy(newentry->nt_domain, domain);
	    }
	  if (debug) syslog(LOG_AUTHPRIV | LOG_DEBUG, "Reauthenticating user %s %s %s %s %d\n", newentry->ux_user, newentry->nt_user, newentry->nt_domain, newentry->pdc_nbname, cache_status);
	  /* this is the actual validation routine call here <------------*/

	  valid=Valid_MapUser(newentry);

	  if (valid==0)
	    {
	      if (found==NULL)
		{
		  /* Add user to cache where no previous entry exists */
		  /* i.e. a new entry */
		  newentry->usertimeout=time(NULL);
		  
		  if (user_cache->head==NULL)
		    {
		      user_cache->head=newentry;
		      user_cache->last=newentry;
		    }
		  else
		    {
		      user_cache->last->next=newentry;
		      user_cache->last=newentry;
		    }
		}
	      else
		{
		  found->usertimeout=time(NULL);
		} 
	    }
	}
      else
	valid=0;
      
      msg.return_code = valid;

      msg.msg_to=msg.msg_fm;
      msg.msg_fm=SERVER;

      memset(msg.username, 0, 16);
      memset(msg.password, 0, 16);
      memset(msg.domain, 0, 16);
      n=msgsnd(mid, &msg, sizeof(msg), 0);
      if (n==-1) {
	perror("server : msgsnd");
	exit(-1);
      }
    }
}

void *sighup_handler(int signum)
{
  smb_readpamconf(primary, backup, domain);
  syslog(LOG_AUTHPRIV | LOG_NOTICE, "Got SIGHUP : reloading conf file : %s %s %s", primary, backup, domain);
  return;
}

void *sigusr2_handler(int signum)
{
  
  syslog(LOG_AUTHPRIV | LOG_NOTICE, "Got SIGUSR2: number of users in cache is %d", user_cache->number_of_cached_users);
  
  return;
}

void *sigterm_handler(int signum)
{
  /* kill the IPC msg queue */
  int n;
  
  n=msgctl(mid ,IPC_RMID, NULL);
  if (n!=0)
    {
      syslog(LOG_AUTHPRIV | LOG_ERR, "Error killing msg queue");
      closelog();
      exit(-1);
    }
  
  closelog();
  exit(0);

}



